from .data import _DataHDU11, _DataHDU22
from .t2 import _T2HDU
from .wavelength import _NW
from .. import utils as _u

import numpy as np

class _VisHDU(_T2HDU):
    """
OI_VIS binary table extension containing the (possibly differential) 
visibility amplitudes, phases, as well as correlated fluxes.
    """
    _EXTNAME = 'OI_VIS'    
    _COLUMNS = [
        ('VISAMP',    True, 'D', (_NW,), None, None, None,
            'visibility amplitude'), 
        ('VISAMPERR', True, 'D', (_NW,), None, None, None,
            'uncertainty on visibility amplitude'),
        ('VISPHI',    True, 'D', (_NW,), None, None, 'deg',
            'phase'), 
        ('VISPHIERR', True, 'D', (_NW,), None, None, 'deg', 
            'uncertainty on phase'),
    ]
    
    @classmethod
    def from_data(cls, *, visamp, visphi, fits_keywords={}, **columns):
        """

Build an OI_VIS table from data.  In the following NWAVE indicates the
number of spectral channels in the configuration and NOBS the number
of observations.

Arguments
---------

version (int)
    version of the OIFITS standard if it cannot be determined from
    context (optional)

date (str)
    date at the start of the first observation (optional, deduced from
    mdj and int_time)
insname (str)
    name of the OI_WAVELENGTH table containing the instrumental setup
arrname (str)         
    name of the OI_ARRAY table containing array properties (optional in
    version 1)
fits_keywords (dict)
    additional FITS keywords (optional)

mjd (float, NOBS)
    MJD of observation
int_time (float, NOBS)
    Integration time
target_id (int, NOBS)
    Target ID in the OI_TARGET table
sta_index (int, NOBS × 2)
    Station IDs of the telescope pair in the OI_ARRAY target (optional in
    version 1)
ucoord, vcoord(float, NOBS)
    (u, v) coordinates of the  baselines

visamp (float, NOBS × NWAVE)
    visibility amplitudes
visamperr (float, NOBS × NWAVE)
    uncertainty on the latter (optional, defaults to 0.)
visphi (float, NOBS × NWAVE)
    visibility phases
visphi (float, NOBS × NWAVE)
    visibility phases (optional, defaults to 0.)
flag (bool, NOBS × NWAVE)
    flag for data (optional, defaults to False)

Arguments in OIFITS1 only
-------------------------

time (float, NOBS)
    seconds since midnight UT on first observation (DATE-OBS i.e. date)

Arguments in OIFITS2 only
-------------------------

corrname (str)
    name of the correlation matrix (optional)
amptyp (str)
    type of amplitude: absolute, differential, correlated flux (optional)
phityp (str)
    type of phase: absolute, differential (optional)
amporder (int)
    order of the polynomial fit for differential amplitudes (optional)
phiorder (int)
    order of the polynomial fit for differential phases (optional)

rvis (float, NOBS × NWAVE)
    real part of the coherent flux (optional)
rviserr (float, NOBS × NWAVE)
    uncertainty on the latter (optional, defaults to 0.)
ivis (float, NOBS × NWAVE)
    imaginary part of the coherent flux (optional)
iviserr (float, NOBS × NWAVE)
    uncertainty on the latter (optional, defaults to 0.)
corrindx_visamp (int, NOBS × NWAVE)
    index in the correlation matrix for the first visamp (optional)  
corrindx_visphi (int, NOBS × NWAVE)
    index in the correlation matrix for the first visphi (optional)
corrindx_rvis (int, NOBS × NWAVE)
    index in the correlation matrix for the first rvis (optional)
corrindx_ivis (int, NOBS × NWAVE)
    index in the correlation matrix for the first ivis (optional)

visref_map (int, NOBS × NWAVE × NWAVE)
    indices of the spectral channels used to determine the differential
    quantities

Additional arguments
--------------------

Any additional keyword argument will be appended as a non-standard FITS 
column with its name prefixed with NS_ 

        """
        columns = dict(visamp=visamp, visphi=visphi,  **columns)

        shape = self._get_columns_shape(**columns)
        for name in ['visamp', 'visphi', 'rvis', 'ivis']:
            if name in columns:
                _u.store_default(columns, f"{name}err", 0., shape)

        return super().from_data(fits_keywords=fits_keywords, **columns)

class VisHDU1(
        _VisHDU,
        _DataHDU11,
      ):
    """

First revision of the OI_VIS binary table, OIFITS v. 1.

    """
    pass

_spos = _u.is_strictpos

class VisHDU2(
        _VisHDU,
        _DataHDU22,
      ):
    """

Second revision of the OI_VIS binary table, OIFITS v. 2.

    """
    _COLUMNS = [
        ('CORRINDX_VISAMP',  False, '1J', (),        _spos, None, None,
            'index on 1st amp. in matching OI_CORR matrix'), 
        ('CORRINDEX_VISPHI', False, '1J', (),        _spos, None, None,
            'index on 1st phase in matching OI_CORR matrix'),
        ('RVIS',             False, 'D', (_NW,),     None,  None, 'any', 
            'real part of correlated flux'),                     
        ('RVISERR',          False, 'D', (_NW,),     None,  None, 'any', 
            'uncertainty on real part of correlated flux'),      
        ('IVIS',             False, 'D', (_NW,),     None,  None, 'any', 
            'imag. part of correlated flux'),                    
        ('IVISERR',          False, 'D', (_NW,),     None,  None, 'any', 
            'uncertainty on imag. part of correlated flux'),
        ('CORRINDX_RVIS',    False, '1J', (),        _spos, None, None,
            'index of 1st RVIS in matching OI_CORR matrix'),
        ('CORRINDEX_IVIS',   False, '1J', (),        _spos, None, None,
            'index of 1st IVIS in matching OI_CORR matrix'),
        ('VISREFMAP',        False, 'L', (_NW,_NW,), None,  None, None,
            'reference channels for differential quantities'),
    ]
    
    _CARDS = [
        ('AMPTYP',   False, _u.is_amptyp2,  None, 
            'amplitude type'),
        ('PHITYP',   False, _u.is_phityp2,  None, 
            'phase is absolute or differential'),
        ('AMPORDER', False, _u.is_posint, 0, 
            'polynomial order in diff. amplitude'),
        ('PHIORDER', False, _u.is_posint, 0, 
            'polynomial order in diff. phase'),
    ]
    
    def get_reference_channel(self, shape='data', flatten=False):

        visrefmap = getattr(self, 'VISREFMAP', 0)
        if visrefmap == 0:
            return super().get_reference_channel('data', False)

        nwave = visrefmap.shape[-1]
        visrefmap = np.array(visrefmap, dtype=object) # allows long int's
        visrefmap = sum(visrefmap[:,:,i] << (i + 1) for i in range(nwave))
        if flatten:
            visrefmap = visrefmap.ravel()
 
        return visrefmap
    
    def get_obs_type(self, name, shape='data', flatten=False):
        
        if name == 'VISAMP':
            typ = self.header.get('AMPTYP', 'N/A')
        elif name == 'VISPHI':
            typ = self.header.get('PHITYP', 'N/A')
        else:
            typ = 'correlated flux'
        return self._resize_data(typ, shape, flatten)
    
    @classmethod
    def from_data(cls, *, visamp, visphi,  

        amptyp=None, phityp=None, amporder=None, phiorder=None, 
        fits_keywords={}, **columns):

        fits_keywords = dict(amptyp=amptyp, phityp=phityp, amporder=amporder,
            phiorder=phiorder, **fits_keywords)
        return super().from_data(cls, visamp=visamp, visphi=visphi, 
                        fits_keywords=fits_keywords, **columns)

    def _verify(self, option='warn'):

        errors = super()._verify(option=option)

        # Format doesn't have consistent units for VISAMP & VISAMPERR
        # Deal with it manually

        amptyp = self.header.get('AMPTYP', 'absolute')
        
        for name in ['VISAMP', 'VISAMPERR']:
            column = self.columns[name]
            if amptyp == 'correlated flux':
                if column.unit is None:
                    err_txt = f"{name} must have unit for {amptyp}"
                    err = self.run_option(option, err_txt)
                    errors.append(err)
            else:
                if column.unit not in ['', None]:
                    err_txt = f"{name} cannot have unit for AMPTYP='{amptyp}'"
                    fix_txt = "Unit removed"
                    def fix(col=column): col.unit = None
                    err = self.run_option(option, err_txt, fix_txt, fix)
                    errors.append(err)
        
        return errors

    def __mod__(self, other):

        h1, h2 = self.header, other.header

        return (super().__mod__(other) and
                h1.get('AMPTYP', '') == h2.get('AMPTYP', '') and
                h1.get('PHITYP', '') == h2.get('PHITYP', '') and
                h1.get('AMPORDER', 0) == h2.get('AMPORDER', 0) and
                h1.get('PHIORDER', 0) == h2.get('PHIORDER', 0))
        
new_vis_hdu = _VisHDU.from_data



from astropy.io.fits.hdu import base as _fitsbase
from astropy.io import fits as _fits
from .. import utils as _u
import copy as _copy

# All Valid HDUs for OIFITS will inherit a _CARDS structured array
# describing the specific FITS keywords in the header
_InheritCardDescription = _u.InheritConstantArray(
    '_CARDS',
    dtype=[
        ('name', 'U32'),     # keyword name
        ('required', bool),  # whether it must be present
        ('test', object),    # func that tests the validity of keyword value
        ('default', object), # default value. None: no default possible
        ('comment', 'U47')   # associated comment
    ]
)

class _ValidHDU(
    _InheritCardDescription,
    _fitsbase._ValidHDU
):

    def copy(self):
        """

Create a duplicate of an OIFITS extension or primary header, with data
and header are copied.

        """
        data = self.data
        if data is not None:
            data = data.copy()
        cp = type(self)(data=data, header=self.header.copy())
        if hasattr(self, '_container'):
            cp._container = self._container
        return cp

    def get_container(self):
        return getattr(self, '_container', None)
    
    def _verify(self, option='warn'):

        errors = super()._verify(option)
       
        header = self.header
        for card in self._CARDS:
            name = card['name']
            if card['required'] or name in header:
                test = card['test']
                default = card['default']
                self.req_cards(name, None, test, default, option, errors)
        
        return errors

    def __ge__(self, other):

        return not self < other

    def __gt__(self, other):

        return not self <= other

    def __lt__(self, other):
    
        return self is not other and self <= other

    def __le__(self, other):

        # A OIFITS primary HDU comes first, then a FITS primary HDU,
        # then an OIFITS extension, and last other FITS extensions.
        if (isinstance(self, _fits.PrimaryHDU) or
            not isinstance(other, _ValidHDU)):
            return True
        
        if isinstance(other, _fits.PrimaryHDU):
            return False

        # OIFITS extensions will be ordered by type first
        order = {'OI_TARGET': 1, 'OI_ARRAY': 2, 'OI_WAVELENGTH': 3,
                 'OI_VIS': 4, 'OI_VIS2': 5, 'OI_T3': 6, 'OI_FLUX': 7,
                 'OI_CORR': 8, 'OI_INSPOL': 9}

        sheader = self.header
        oheader = other.header
        sext = sheader.get('EXTNAME', '')
        oext = oheader.get('EXTNAME', '')
        sorder = order.get(sext, 99999)
        oorder = order.get(oext, 99999)

        if sorder != oorder:
            return sorder < oorder

        # ordering uses reference names, date of observation and, in the
        # last resort, memory address.
        
        arr1 = sheader.get('ARRNAME', '')
        arr2 = sheader.get('ARRNAME', '')
        ins1 = sheader.get('INSNAME', '')
        ins2 = oheader.get('INSNAME', '')
        corr1 = sheader.get('CORRNAME', '')
        corr2 = oheader.get('CORRNAME', '')
        mjd1 = min(getattr(self.data, 'MJD', [0]))
        mjd2 = min(getattr(other.data, 'MJD', [0]))
        id1 = id(self)
        id2 = id(other)

        return (arr1, ins1, corr1, mjd1, id1) <  (arr2, ins2, corr2, mjd2, id2)

    def to_version(self, version):
       
        if self._OI_VER == version:
            return self

        newclass = self.get_class(version=version) 
        newobj =  newclass(data=self.data, header=self.header)

        return newobj 

    @classmethod
    def get_class(cls, version=2):
        """Get the class corresponding to OIFITS version"""
        
        if hasattr(cls, '_OI_VER'):
            cls = cls.__base__
        
        subclasses = cls.__subclasses__()
        for newcls in subclasses:
            if getattr(newcls, '_OI_VER', None) == version:
                break
        else:
            error = f"Could not find a class maching {version=}"
            raise RuntimeError(error)

        return newcls

class _OIFITS1HDU(_ValidHDU):
    _OI_VER = 1

class _OIFITS2HDU(_ValidHDU):
    _OI_VER = 2




#! /usr/bin/env python3 

import setuptools
import os
import re

PACKAGE = setuptools.find_packages(exclude=('tests',))[0]

def get_readme():
    with open("README.md", "r") as fh:
        text = fh.read()
    return text

def read_init(splitlines=False):
    here = os.path.abspath(os.path.dirname(__file__))
    filename = os.path.join(here, PACKAGE, '__init__.py')
    with open(filename) as fh:
        lines = fh.read()
    if splitlines:
        lines = lines.splitlines()
    return lines

def get_version():
    lines = read_init(splitlines=True)
    for line in lines:
        match = re.search('__version__\s*=\s*([\'"])(.*)\\1', line)
        if match:
            return match.groups()[1]
    raise RuntimeError("Unable to find version string.")

# Python package install
 
setuptools.setup(
    name='pyoifits',
    version=get_version(),
    packages=setuptools.find_packages(),
    license='LICENSE.txt',
    author="Régis Lachaume",
    author_email="regis.lachaume@gmail.com",
    description='OIFITS (Data Exchange Standard for Optical Interferometry, A&A 597, A8, 2017)',
    long_description=get_readme(),
    long_description_content_type='text/markdown',
    classifiers=[
        "Programming Language :: Python :: 3",
        "Development Status :: 3 - Alpha",
        "Operating System :: OS Independent",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Astronomy",
        "License :: Public Domain",
    ],
    python_requires='>=3.8',
    install_requires=[
        "astropy>=4.0", 
        "numpy>=1.17",
        "scipy>=1.5",
        "astroquery>=0.4",
    ],
)



"""
A package to read, write and manipulate files containing the Optical 
Interferometry Data Exchange standard format defined by
* OIFITS 2: Gilles Duvert, John Young & Christian Hummel (2017),
  Astronomy & Astrophysics, 597, A8, doi:10.1051/0004-6361/201526405, 
  arXiv:1510.04556 
* OIFITS 1: T.A. Pauls et al. (2005), Publications of the Astronomical 
  Society of the Pacific 117, 1255, doi:10.1086/444523, 
  arXiv:astro-ph/050818

Typical use

>>> from pyoifits import open as oifits_open
>>> hdulist = oifits_open('filename.fits')

"""

from .oifits import *

__version__ = "0.4.3"



"""
Implementation of the OI_ARRAY binary table extension.
"""

from .table import _OITableHDU, _OITableHDU11, _OITableHDU22
from .referenced import _Referenced
from astropy.coordinates import EarthLocation as _EarthLocation

from .. import utils as _u
from .. import coo as _coo

import numpy as _np

class _ArrayHDUBase(_OITableHDU):

    def _get_array_field(self, name, shape='none', flatten=False,
            concatenate=False, default=None):

        refhdu = self.get_arrayHDU()

        if refhdu is None:
            if default is None:
                return None
            val = default
        else:
            val = self._xmatch(refhdu, 'STA_INDEX', 
                            name=name, concatenate=concatenate)

        return self._resize_data(val, shape, flatten)

    def get_sta_name(self, shape='none', flatten=False):
        return self._get_array_field('STA_NAME', shape, flatten)

    def get_staxyz(self):
        return self._get_array_field('STAXYZ')

    def get_staenu(self):
        """

Determine the (East, North, up) coordinates of each station with respect
with the centre of the array

Arguments
---------

None

Returns
-------

uvw (float × NOBS × NSTA × 3)
    (u, v, w) coordinates for each observation and each station. NOBS
    is the number of observations (rows) in the table and NSTA the number
    of stations (NSTA = 3 for OI_T3, NSTA = 2 for OI_VIS and OI_VIS2,
    NSTA = 1 for OI_FLUX in uncalibrated mode)

Raises
------

NotImplementedError 
    The array is not defined in the GEOCENTRIC frame

        """
        xyz = self.get_staxyz()
        frame = self.header.get('FRAME', 'GEOCENTRIC')

        if frame != 'GEOCENTRIC':
            txt = f"(E, N, U) coordinates not implemented for frame {frame}"
            raise NotImplementedErrror(txt)

        # (X, Y, Z)_ITRS -> (E, N, U)_local 
        loc = self.get_location()
        lon = loc.lon.to_value('rad')
        lat = loc.lat.to_value('rad')
        staune = _u.rotation3d(xyz, 'zy', [-lon, lat], degrees=False)
        staenu = _np.roll(staune, -1, axis=-1)
       
        return staenu
 
    def get_sta_config(self, shape='none', flatten=False, default=None):
        return self._get_array_field('STA_NAME', shape, flatten,
                    concatenate=True, default=default)

    def get_tel_name(self, shape='none', flatten=False):
        return self._get_array_field('TEL_NAME', shape, flatten)

    def get_tel_config(self, shape='none', flatten=False, default=None):
        return self._get_array_field('TEL_NAME', shape, flatten,
                    concatenate=True, default=default)

    def get_arrname(self, shape='none', flatten=False, default=None):
        arrname = self.header.get('ARRNAME', default)
        if not arrname:
            return None
        arrname = self._resize_data(arrname, shape, flatten)
        return arrname

    def get_location(self):
        
        arrayHDU = self.get_arrayHDU()
        header = arrayHDU.header
        xyz = [header.get('ARRAY' + var, 0) for var in ['X', 'Y', 'Z']]
        loc = _EarthLocation.from_geocentric(*xyz, unit="m")

        return loc

    def get_arrayHDU(self):
        if self.header['EXTNAME'] == 'OI_ARRAY':
            return self
        container = self.get_container()
        if container is None:
            return None
        return container.get_arrayHDU(arrname=self.get_arrname())

    def _verify(self, option='warn'):

        errors = super()._verify(option)
      
        # If no ARRNAME and STA_INDEX, fine (optional), but if
        # only one of them is present, there's a problem.
        sta_index = getattr(self, 'STA_INDEX', None)
        arrname = self.header.get('ARRNAME', None) 
        
        if not arrname:
            if sta_index is not None and not isinstance(self, _ArrayHDU):
                err_text = "STA_INDEX column present with no ARRNAME keyword"
                err = self.run_option(option, err_text, fixable=False)
                errors.appen(err)
            return errors
         
        if sta_index is None:
            err_text = "ARRNAME is given but not STA_INDEX column"
            err = self.run_option(option, err_text, fixable=False)
            errors.append(err)
            return errors

        # check it has the corresponding arrayHDU
        refhdu = self.get_arrayHDU() 
        if refhdu is None:
            err_text = "ArrayHDU with ARRNAME={arrname} not found"
            err = self.run_option(option, err_text, fixable=False)
            errors.append(err)
            return errors

        # We don't report on i <= 0 unreferenced indices: they are
        # already spotted as not valid by column verification 
        ref_index = refhdu.STA_INDEX
        sta_index = _np.unique(sta_index)
        missing = [str(i) for i in sta_index if i > 0 and i not in ref_index]
        if len(missing):
            m = ', '.join(missing)
            err_text = f"'STA_INDEX' not referenced in ArrayHDU: {m}"
            err = self.run_option(option, err_text, fixable=False)
            errors.append(err)

        return errors

    def __ge__(self, other):
        return (isinstance(self, _OITableHDU) and
                h1.get_arrname() == other.get('ARRNAME', ''))
    def __gt__(self, other):
        return self is not other and self >= other

class _MayHaveArrayHDU(_ArrayHDUBase):
    _CARDS = [('ARRNAME', False, _u.is_nonempty, None, 
        'Name of telescope array for cross-reference')]

class _MustHaveArrayHDU(_ArrayHDUBase):
    _CARDS = [('ARRNAME', True, _u.is_nonempty, None, 
        'Name of telescope array for cross-reference')] 

class _ArrayHDU(_MustHaveArrayHDU,_Referenced):
    
    _EXTNAME = 'OI_ARRAY'
    _REFERENCE_KEY = 'ARRNAME'
    _CARDS = [
        ('ARRAYX', True, _u.is_num, None, 'X coordinate of array centre'),
        ('ARRAYY', True, _u.is_num, None, 'Y coordinate of array centre'),
        ('ARRAYZ', True, _u.is_num, None, 'Z coordinate of array centre'),
    ]
    _COLUMNS = [
        ('TEL_NAME',  True, '16A', (),  None,            None, None,
                'telescope name'), 
        ('STA_NAME',  True, '16A', (),  None,            None, None,
                'station name'), 
        ('STA_INDEX', True, '1I',  (),  _u.is_strictpos, None, None,
                'station index for cross-reference'),
        ('DIAMETER',  True, '1E', (),   _u.is_strictpos, None, "m",
                'effective diameter of the aperture'),
        ('STAXYZ',    True, '3E', (3,), None,            None, "m",
                'station coordinates relative to array centre'),
    ]
    
    def _get_ins(self):
        name = self.arrname()
        m = re.match('[A-Za-z_]{1,20}', name)
        if m:
            return m.group()
        return name[0:20] 

    def _verify(self, option='warn'):

        errors = super()._verify(option)
        
        sta_index = self.STA_INDEX
        if len(_np.unique(sta_index)) == len(sta_index):
            return errors

        err_text = f"Repeated STA_INDEX in {type(self).__name__}"
        err = self.run_option(option, err_text, fixable=False)
        errors.append(err)

        return errors

    def __add__(self, other):

        return self.merge(other)

    def merge(self, *others):

        norm = _np.linalg.norm
        dist_max = self.get_container()._merge_station_distance

        def eq(x, y):
            return ((x['STA_NAME'] == y['STA_NAME']) and
                    (x['TEL_NAME'] == y['TEL_NAME']) and
                    (norm(x['STAXYZ'] - y['STAXYZ']) <= dist_max))
        
        return self._merge_helper(*others, id_name='STA_INDEX', equality=eq)

    def __mod__(self, other):

        h1, h2 = self.header, other.header
        dist_max = self.get_container()._merge_array_distance
        return (self & other and
                h1['ARRNAME'] == h2['ARRNAME'] and
                h1['FRAME'] == h2['FRAME'] and
                abs(h1['ARRAYX'] - h2['ARRAYX']) <= dist_max and
                abs(h1['ARRAYY'] - h2['ARRAYY']) <= dist_max and
                abs(h1['ARRAYZ'] - h2['ARRAYZ']) <= dist_max)

    def is_referred_to_by(self, other):
        return (not isinstance(other, _ArrayHDU) and
                isinstance(other, _ArrayHDUBase) and
                self.get_arrname() == other.get_arrname())

    @classmethod
    def from_data(cls, *, arrname, version=2, 
            arrayxyz=None, lat=None, lon=None, alt=None, 
            frame='GEOCENTRIC', ellipsoid='WGS84',
            tel_name=None, sta_name=None, sta_index=None, diameter=None, 
            staxyz=None, staenu=None,
            fits_keywords={}, **columns):
        """

Build an OI_ARRAY extension from data. In the following, NSTA is the
number of stations.

Arguments
---------

version:
    OIFITS version if it cannot be determined from context (optional)
ellipsoid (str)
    Geodetic coordinate system (optional, default is WGS84)
calibrated (bool)
    Whether flux are calibrated or uncalibrated (measured at each station)

arrname (str)
    name of the array
frame (str)
    frame (optional, default is GEOCENTRIC)
fits_keywords (dict)
    additional FITS keywords (optional)
arrayxyz (float, 3) 
or lat, lon, alt (float):
    Coordinates of the centre of the array. 
    * arrayxyz uses the geocentric frame in metres.
    * lat, lon, alt are the geodetic coordinates in degrees and metres.
    Only for GEOCENTRIC frame.

sta_index (int, NSTA)
    Station index for cross-reference
tel_name (str, NSTA)
    Names of the telescopes
sta_name (str, NSTA)
    Names of the stations
diameter (float, NSTA)
    Effective aperture of the telescopes
staxyz (float, NSTA × 3)
or staenu (float, NSTA × 3)
    Coordinates of the stations relative to array centre. staenu uses
    the local frame (East-North-up) and sta_xyz the specified frame
    (Cartesian geocentric if GEOCENTRIC, local frame if SKY). In metres.

Arguments in OIFITS2 only
-------------------------

fov (float)
    field of view in arcsec
fovtype (str)
    field of view type is FWHM or RADIUS 

Additional arguments
--------------------

Any additional keyword argument will be appended as a non-standard FITS 
column with its name prefixed with NS_ 

        """
        geodetic = lon is not None and lat is not None and alt is not None
        if geodetic:
            if frame == 'GEOCENTRIC':
                loc = _EarthLocation.from_geodetic(lon, lat, alt, ellipsoid)
                arrayxyz = [loc.x.value, loc.y.value, loc.z.value]
        if frame == 'SKY':
            arrayxyz = [0., 0., 0.]

        if lon is not None:
            fits_keywords['LON-OBS'] = (lon, 'longitude of array centre (deg)')
        if lat is not None:
            fits_keywords['LAT-OBS'] = (lat, 'latitude of array centre (deg)')
        if alt is not None:
            fits_keywords['ALT-OBS'] = (alt, 'altitude of array centre (m)')
   
        if staenu is not None:
            if frame == 'SKY':
                staxyz = staenu
            else:
                if not geodetic:
                    loc = _EarthLocation.from_geocentric(*arrayxyz, unit="m")
                    loc = loc.to_geodetic(ellipsoid=ellipsoid)
                    lon = loc.lon.value
                    lat = loc.lat.value
                # (E, N, U) -> (X, Y, Z)_ITRS transform
                staune = _np.roll(staenu, 1, axis=-1)
                staxyz = _u.rotation3d(staune, 'yz', [-lat, lon], degrees=True)
        if sta_index is None:
            sta_index = list(range(1, len(tel_name) + 1))
        
        fits_keywords = dict(arrname=arrname, frame=frame,
            arrayx=arrayxyz[0], arrayy=arrayxyz[1], arrayz=arrayxyz[2],
            **fits_keywords)
   
        columns = dict(tel_name=tel_name, sta_name=sta_name, 
            sta_index=sta_index, diameter=diameter, staxyz=staxyz,
            **columns)
        
        return super().from_data(version=version, 
                    fits_keywords=fits_keywords, **columns)

            

class ArrayHDU1(
        _ArrayHDU,
        _OITableHDU11, # OFITS1, rev. 1
      ):
    """

First revision of the OI_ARRAY binary table, OIFITS v. 1

    """
    _CARDS = [('FRAME', True, _u.is_frame1,  'GEOCENTRIC', 
        'coordinate frame for array centre and stations')]

class ArrayHDU2(
        _ArrayHDU,
        _OITableHDU22, # OIFITS2, rev. 2
      ):
    """

Second revision of the OI_ARRAY binary table, OIFITS v. 2

    """
    _CARDS = [('FRAME', True, _u.is_frame2,  'GEOCENTRIC', 
        'coordinate frame for array centre and stations')]
    _COLUMNS = [
        ('FOV',     False, '1D', (), _u.is_strictpos, None, "arcsec",
                'photometric field of view'), 
        ('FOVTYPE', False, '6A', (), _u.is_fovtype2,  None, None,
                'definition of the field of view')
    ]

new_array_hdu = _ArrayHDU.from_data
